/*
(C) 2012 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>  /* for sei() */
#include <util/delay.h>     /* for _delay_ms() */

#include <avr/pgmspace.h>   /* required by usbdrv.h */
#include "usbdrv.h"
//#include "usbdrv/usbdrv.c"
#include "oddebug.h"        /* This is also an example for using debug macros */

static uchar    reportBuffer[2];    /* buffer for HID reports */
static uchar    idleRate;           /* in 4 ms units */

PROGMEM char usbHidReportDescriptor[35] = { /* USB report descriptor */
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};

#define KEY_A       4
#define KEY_B       5
#define KEY_C       6
#define KEY_D       7
#define KEY_E       8
#define KEY_F       9
#define KEY_1       30
#define KEY_2       31
#define KEY_3       32
#define KEY_4       33
#define KEY_5       34
#define KEY_6       35
#define KEY_7       36
#define KEY_8       37
#define KEY_9       38
#define KEY_0       39
#define KEY_ENTER	0x28

//The button received is used as an index into this array, and the indicated
//key, followed by an enter, is sent when the button is pressed. This array
//is sorted randomly because I grabbed random wires when soldering the buttons
//together :)
PROGMEM char btnToKey[]={
	KEY_1, //0
	KEY_C, //1
	KEY_7, //2
	KEY_5, //3
	KEY_F, //4
	KEY_2, //5
	KEY_4, //6
	KEY_A, //7
	KEY_8, //8
	KEY_D, //9
	KEY_6, //A
	KEY_B, //B
	KEY_3, //C
	KEY_9, //D
	KEY_B, //E
	KEY_E  //F
};

usbMsgLen_t usbFunctionSetup(uchar data[8]) {
	usbRequest_t    *rq = (void *)data;
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
           usbMsgPtr = (void *)&reportBuffer;
	       return sizeof(reportBuffer);
        } else if(rq->bRequest == USBRQ_HID_GET_IDLE){
            usbMsgPtr = &idleRate;
            return 1;
        } else if(rq->bRequest == USBRQ_HID_SET_IDLE){
            idleRate = rq->wValue.bytes[1];
        }
    }else{
        /* no vendor specific requests implemented */
    }
    return 0;   /* default for not implemented requests: return no data back to host */
}

/* ------------------------------------------------------------------------- */

//Select a row in the keyboard/led matrix
static void selectRow(char x) {
	unsigned char bit=(0x10<<x);
	unsigned char b;
	DDRB|=0xF0;
	b=PORTB&0xF;
	b|=(~bit)&0xf0;
	PORTB=b; //make bit low
}

//Send data to the column. An '1' in X will light up the LED in the corresponding
//position.
static void setCol(char x) {
	DDRB|=0x0F;
	PORTB=(PORTB&0xF0)|(x);
}

//Turn on pull-up and read the button values of a column of buttons. An '1'
//in the returned value will indicate that button is pressed.
static char readCol(void) {
	char r;
	DDRB&=~(0xF);
	PORTB|=0xF;
	_delay_us(50);
	r=(~PINB)&0xF;
	PORTB&=~0xF;
	return r;
}

//Read out all the 16 buttons and stuff their status as bit values in the
//returned unsigned int.
static unsigned int getButtons(void) {
	unsigned int ret=0;
	selectRow(0);
	ret=readCol();
	selectRow(1);
	ret|=readCol()<<4;
	selectRow(2);
	ret|=readCol()<<8;
	selectRow(3);
	ret|=readCol()<<12;
	return ret;
}

//Do one cycle of multiplexing the value indicated by the bits in the 'leds'-argument
static void lightLeds(unsigned int leds) {
	char b;
	int x;
	for (x=0; x<4; x++) {
		b=leds&0xF;
		leds>>=4;
		selectRow(x);
		setCol(b);
		_delay_ms(1);
	}
}

//Idle for the amount of time in ms. Also multiplexes the led value to the
//leds, and kicks the watchdog and usb.
static void idle(int ms, unsigned int leds) {
	int x=0;
	for (x=0; x<ms/4; x++) {
		lightLeds(leds);
        wdt_reset();
		usbPoll();
	}
}


//States:
// IDLE -> wait for button press
// SENDDIGIT -> send key corresponding to the pressed button
// RELDIGIT -> release that key
// SENDENTER -> send an enter keypress
// RELENTER -> release the enter key
// LIGHT -> light up the LED under the button for 3 seconds and resume to the IDLE
// state.
#define STATE_IDLE		0
#define STATE_SENDDIGIT	1
#define STATE_RELDIGIT	2
#define STATE_SENDENTER 3
#define STATE_RELENTER  4
#define STATE_LIGHT		5


int __attribute__((noreturn)) main(void) {
	uchar   i;
	uchar state=0;
	unsigned int btn;
	int x, keyNo;
    wdt_enable(WDTO_1S);
    /* Even if you don't use the watchdog, turn it off here. On newer devices,
     * the status of the watchdog (on/off, period) is PRESERVED OVER RESET!
     */
    /* RESET status: all port bits are inputs without pull-up.
     * That's the way we need D+ and D-. Therefore we don't need any
     * additional hardware initialization.
     */
    odDebugInit();
    usbInit();
    usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
    i = 0;
    while(--i){             // fake USB disconnect for > 250 ms
        wdt_reset();
        _delay_ms(1);
    }

    usbDeviceConnect();
    sei();
    for(;;){                /* main event loop */
		if (state==STATE_IDLE) {
			btn=getButtons();
			if (btn!=0) {
				//Idling, but a button has been pressed! The button will be
				//indicated by the position of the '1' in btn.
				state=STATE_SENDDIGIT;
				//Find out what button has been pressed.
				for (x=0; x<16; x++) {
					if (btn&(1<<x)) keyNo=x;
				}
			}
		}
		if (state==STATE_LIGHT) {
			//Sent the keys, now wait for 3 seconds while lighting the 
			//LED of the button that got pressed. Ignore all other buttons.
			idle(3000, btn);
			state=STATE_IDLE;
		}
        wdt_reset();
        usbPoll();
        if(usbInterruptIsReady()){
			reportBuffer[0]=0;
			if (state==STATE_RELDIGIT || state==STATE_RELENTER) {
				//Release whatever was pressed previously.
				reportBuffer[1]=0; //no key
				usbSetInterrupt((void *)&reportBuffer, sizeof(reportBuffer));
				state++;
			} else if (state==STATE_SENDDIGIT) {
				//Send the digit corresponding to button keyNo.
				reportBuffer[1]=pgm_read_byte(&btnToKey[keyNo]);
				usbSetInterrupt((void *)&reportBuffer, sizeof(reportBuffer));
				state++;
			} else if (state==STATE_SENDENTER) {
				//Send an enter key press.
				reportBuffer[1]=KEY_ENTER;
				usbSetInterrupt((void *)&reportBuffer, sizeof(reportBuffer));
				state++;
			}
        }
    }
}

/* ------------------------------------------------------------------------- */
